@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2010 The Music Player Daemon Project
+ * Copyright (C) 2003-2011 The Music Player Daemon Project
  * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
@@ -20,6 +20,8 @@
 #include "config.h"
 #include "decoder_api.h"
 #include "audio_check.h"
+#include "ffmpeg_metadata.h"
+#include "tag_handler.h"
 
 #include <glib.h>
 
@@ -32,11 +34,6 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
-#ifdef OLD_FFMPEG_INCLUDES
-#include <avcodec.h>
-#include <avformat.h>
-#include <avio.h>
-#else
 #include <libavcodec/avcodec.h>
 #include <libavformat/avformat.h>
 #include <libavformat/avio.h>
@@ -46,13 +43,10 @@
 #if LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(51,5,0)
 #include <libavutil/dict.h>
 #endif
-#endif
 
 #undef G_LOG_DOMAIN
 #define G_LOG_DOMAIN "ffmpeg"
 
-#ifndef OLD_FFMPEG_INCLUDES
-
 static GLogLevelFlags
 level_ffmpeg_to_glib(int level)
 {
@@ -84,12 +78,6 @@ mpd_ffmpeg_log_callback(G_GNUC_UNUSED void *ptr, int level,
 	}
 }
 
-#endif /* !OLD_FFMPEG_INCLUDES */
-
-#ifndef AV_VERSION_INT
-#define AV_VERSION_INT(a, b, c) (a<<16 | b<<8 | c)
-#endif
-
 struct mpd_ffmpeg_stream {
 	struct decoder *decoder;
 	struct input_stream *input;
@@ -119,7 +107,7 @@ mpd_ffmpeg_stream_seek(void *opaque, int64_t pos, int whence)
 	if (whence == AVSEEK_SIZE)
 		return stream->input->size;
 
-	if (!input_stream_seek(stream->input, pos, whence, NULL))
+	if (!input_stream_lock_seek(stream->input, pos, whence, NULL))
 		return -1;
 
 	return stream->input->offset;
@@ -189,9 +177,7 @@ mpd_ffmpeg_stream_close(struct mpd_ffmpeg_stream *stream)
 static bool
 ffmpeg_init(G_GNUC_UNUSED const struct config_param *param)
 {
-#ifndef OLD_FFMPEG_INCLUDES
 	av_log_set_callback(mpd_ffmpeg_log_callback);
-#endif
 
 	av_register_all();
 	return true;
@@ -248,6 +234,21 @@ time_to_ffmpeg(double t, const AVRational time_base)
 }
 
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(53,25,0)
+
+static void
+copy_interleave_frame2(uint8_t *dest, uint8_t **src,
+		       unsigned nframes, unsigned nchannels,
+		       unsigned sample_size)
+{
+	for (unsigned frame = 0; frame < nframes; ++frame) {
+		for (unsigned channel = 0; channel < nchannels; ++channel) {
+			memcpy(dest, src[channel] + frame * sample_size,
+			       sample_size);
+			dest += sample_size;
+		}
+	}
+}
+
 /**
  * Copy PCM data from a AVFrame to an interleaved buffer.
  */
@@ -268,11 +269,10 @@ copy_interleave_frame(const AVCodecContext *codec_context,
 
 	if (av_sample_fmt_is_planar(codec_context->sample_fmt) &&
 	    codec_context->channels > 1) {
-		for (int i = 0, channels = codec_context->channels;
-		     i < channels; i++) {
-			memcpy(buffer, frame->extended_data[i], plane_size);
-			buffer += plane_size;
-		}
+		copy_interleave_frame2(buffer, frame->extended_data,
+				       frame->nb_samples,
+				       codec_context->channels,
+				       av_get_bytes_per_sample(codec_context->sample_fmt));
 	} else {
 		memcpy(buffer, frame->extended_data[0], data_size);
 	}
@@ -287,7 +287,7 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 		   AVCodecContext *codec_context,
 		   const AVRational *time_base)
 {
-	if (packet->pts != (int64_t)AV_NOPTS_VALUE)
+	if (packet->pts >= 0 && packet->pts != (int64_t)AV_NOPTS_VALUE)
 		decoder_timestamp(decoder,
 				  time_from_ffmpeg(packet->pts, *time_base));
 
@@ -318,20 +318,33 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 	       cmd == DECODE_COMMAND_NONE) {
 		int audio_size = buffer_size;
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(53,25,0)
-		AVFrame frame;
+
+		AVFrame *frame = avcodec_alloc_frame();
+		if (frame == NULL) {
+			g_warning("Could not allocate frame");
+			break;
+		}
+
 		int got_frame = 0;
 		int len = avcodec_decode_audio4(codec_context,
-						&frame, &got_frame,
+						frame, &got_frame,
 						&packet2);
 		if (len >= 0 && got_frame) {
 			audio_size = copy_interleave_frame(codec_context,
-							   &frame,
+							   frame,
 							   aligned_buffer,
 							   buffer_size);
 			if (audio_size < 0)
 				len = audio_size;
 		} else if (len >= 0)
 			len = -1;
+
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(54, 28, 0)
+		avcodec_free_frame(&frame);
+#else
+		av_freep(&frame);
+#endif
+
 #elif LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52,25,0)
 		int len = avcodec_decode_audio3(codec_context,
 						aligned_buffer, &audio_size,
@@ -344,7 +357,7 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 
 		if (len < 0) {
 			/* if error, we skip the frame */
-			g_message("decoding failed\n");
+			g_message("decoding failed, frame skipped\n");
 			break;
 		}
 
@@ -366,13 +379,20 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 	return cmd;
 }
 
+#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(52, 94, 1)
+#define AVSampleFormat SampleFormat
+#endif
+
+G_GNUC_CONST
 static enum sample_format
-ffmpeg_sample_format(G_GNUC_UNUSED const AVCodecContext *codec_context)
+ffmpeg_sample_format(enum AVSampleFormat sample_fmt)
 {
-#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(51, 41, 0)
-	switch (codec_context->sample_fmt) {
+	switch (sample_fmt) {
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52, 94, 1)
 	case AV_SAMPLE_FMT_S16:
+#if LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(51,17,0)
+	case AV_SAMPLE_FMT_S16P:
+#endif
 #else
 	case SAMPLE_FMT_S16:
 #endif
@@ -380,20 +400,35 @@ ffmpeg_sample_format(G_GNUC_UNUSED const AVCodecContext *codec_context)
 
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52, 94, 1)
 	case AV_SAMPLE_FMT_S32:
+#if LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(51,17,0)
+	case AV_SAMPLE_FMT_S32P:
+#endif
 #else
 	case SAMPLE_FMT_S32:
 #endif
 		return SAMPLE_FORMAT_S32;
 
-	default:
-		g_warning("Unsupported libavcodec SampleFormat value: %d",
-			  codec_context->sample_fmt);
-		return SAMPLE_FORMAT_UNDEFINED;
-	}
-#else
-	/* XXX fixme 16-bit for older ffmpeg (13 Aug 2007) */
-	return SAMPLE_FORMAT_S16;
+#if LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(51,17,0)
+	case AV_SAMPLE_FMT_FLTP:
+		return SAMPLE_FORMAT_FLOAT;
 #endif
+
+	default:
+		break;
+	}
+
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52, 94, 1)
+	char buffer[64];
+	const char *name = av_get_sample_fmt_string(buffer, sizeof(buffer),
+						    sample_fmt);
+	if (name != NULL)
+		g_warning("Unsupported libavcodec SampleFormat value: %s (%d)",
+			  name, sample_fmt);
+	else
+#endif
+		g_warning("Unsupported libavcodec SampleFormat value: %d",
+			  sample_fmt);
+	return SAMPLE_FORMAT_UNDEFINED;
 }
 
 static AVInputFormat *
@@ -406,7 +441,8 @@ ffmpeg_probe(struct decoder *decoder, struct input_stream *is)
 
 	unsigned char *buffer = g_malloc(BUFFER_SIZE);
 	size_t nbytes = decoder_read(decoder, is, buffer, BUFFER_SIZE);
-	if (nbytes <= PADDING || !input_stream_seek(is, 0, SEEK_SET, NULL)) {
+	if (nbytes <= PADDING ||
+	    !input_stream_lock_seek(is, 0, SEEK_SET, NULL)) {
 		g_free(buffer);
 		return NULL;
 	}
@@ -503,11 +539,16 @@ ffmpeg_decode(struct decoder *decoder, struct input_stream *input)
 		return;
 	}
 
+	const enum sample_format sample_format =
+		ffmpeg_sample_format(codec_context->sample_fmt);
+	if (sample_format == SAMPLE_FORMAT_UNDEFINED)
+		return;
+
 	GError *error = NULL;
 	struct audio_format audio_format;
 	if (!audio_format_init_checked(&audio_format,
 				       codec_context->sample_rate,
-				       ffmpeg_sample_format(codec_context),
+				       sample_format,
 				       codec_context->channels, &error)) {
 		g_warning("%s", error->message);
 		g_error_free(error);
@@ -588,72 +629,24 @@ ffmpeg_decode(struct decoder *decoder, struct input_stream *input)
 	mpd_ffmpeg_stream_close(stream);
 }
 
-#if LIBAVFORMAT_VERSION_INT >= ((52<<16)+(31<<8)+0)
-typedef struct ffmpeg_tag_map {
-	enum tag_type type;
-	const char *name;
-} ffmpeg_tag_map;
-
-static const ffmpeg_tag_map ffmpeg_tag_maps[] = {
-#if LIBAVFORMAT_VERSION_INT < ((52<<16)+(50<<8))
-	{ TAG_ARTIST,            "author" },
-	{ TAG_DATE,              "year" },
-#endif
-	{ TAG_ARTIST_SORT,       "author-sort" },
-	{ TAG_ALBUM_ARTIST,      "album_artist" },
-	{ TAG_ALBUM_ARTIST_SORT, "album_artist-sort" },
-
-	/* sentinel */
-	{ TAG_NUM_OF_ITEM_TYPES, NULL }
-};
-
-#if LIBAVFORMAT_VERSION_INT < AV_VERSION_INT(53,1,0)
-#define AVDictionary AVMetadata
-#define AVDictionaryEntry AVMetadataTag
-#define av_dict_get av_metadata_get
-#endif
-
-static void
-ffmpeg_copy_metadata(struct tag *tag, enum tag_type type,
-		     AVDictionary *m, const char *name)
-{
-	AVDictionaryEntry *mt = NULL;
-
-	while ((mt = av_dict_get(m, name, mt, 0)) != NULL)
-		tag_add_item(tag, type, mt->value);
-}
-
-static void
-ffmpeg_copy_dictionary(struct tag *tag, AVDictionary *dict)
-{
-	for (unsigned i = 0; i < TAG_NUM_OF_ITEM_TYPES; ++i)
-		ffmpeg_copy_metadata(tag, i,
-				     dict, tag_item_names[i]);
-
-	for (const struct ffmpeg_tag_map *i = ffmpeg_tag_maps;
-	     i->name != NULL; ++i)
-		ffmpeg_copy_metadata(tag, i->type, dict, i->name);
-}
-
-#endif
-
 //no tag reading in ffmpeg, check if playable
-static struct tag *
-ffmpeg_stream_tag(struct input_stream *is)
+static bool
+ffmpeg_scan_stream(struct input_stream *is,
+		   const struct tag_handler *handler, void *handler_ctx)
 {
 	AVInputFormat *input_format = ffmpeg_probe(NULL, is);
 	if (input_format == NULL)
-		return NULL;
+		return false;
 
 	struct mpd_ffmpeg_stream *stream = mpd_ffmpeg_stream_open(NULL, is);
 	if (stream == NULL)
-		return NULL;
+		return false;
 
 	AVFormatContext *f = NULL;
 	if (mpd_ffmpeg_open_input(&f, stream->io, is->uri,
 				  input_format) != 0) {
 		mpd_ffmpeg_stream_close(stream);
-		return NULL;
+		return false;
 	}
 
 #if LIBAVFORMAT_VERSION_INT >= AV_VERSION_INT(53,6,0)
@@ -669,49 +662,22 @@ ffmpeg_stream_tag(struct input_stream *is)
 		av_close_input_stream(f);
 #endif
 		mpd_ffmpeg_stream_close(stream);
-		return NULL;
+		return false;
 	}
 
-	struct tag *tag = tag_new();
+	if (f->duration != (int64_t)AV_NOPTS_VALUE)
+		tag_handler_invoke_duration(handler, handler_ctx,
+					    f->duration / AV_TIME_BASE);
 
-	tag->time = f->duration != (int64_t)AV_NOPTS_VALUE
-		? f->duration / AV_TIME_BASE
-		: 0;
-
-#if LIBAVFORMAT_VERSION_INT >= ((52<<16)+(31<<8)+0)
 #if LIBAVFORMAT_VERSION_INT < AV_VERSION_INT(52,101,0)
 	av_metadata_conv(f, NULL, f->iformat->metadata_conv);
 #endif
 
-	ffmpeg_copy_dictionary(tag, f->metadata);
+	ffmpeg_scan_dictionary(f->metadata, handler, handler_ctx);
 	int idx = ffmpeg_find_audio_stream(f);
 	if (idx >= 0)
-		ffmpeg_copy_dictionary(tag, f->streams[idx]->metadata);
-#else
-	if (f->author[0])
-		tag_add_item(tag, TAG_ARTIST, f->author);
-	if (f->title[0])
-		tag_add_item(tag, TAG_TITLE, f->title);
-	if (f->album[0])
-		tag_add_item(tag, TAG_ALBUM, f->album);
-
-	if (f->track > 0) {
-		char buffer[16];
-		snprintf(buffer, sizeof(buffer), "%d", f->track);
-		tag_add_item(tag, TAG_TRACK, buffer);
-	}
-
-	if (f->comment[0])
-		tag_add_item(tag, TAG_COMMENT, f->comment);
-	if (f->genre[0])
-		tag_add_item(tag, TAG_GENRE, f->genre);
-	if (f->year > 0) {
-		char buffer[16];
-		snprintf(buffer, sizeof(buffer), "%d", f->year);
-		tag_add_item(tag, TAG_DATE, buffer);
-	}
-
-#endif
+		ffmpeg_scan_dictionary(f->streams[idx]->metadata,
+				       handler, handler_ctx);
 
 #if LIBAVFORMAT_VERSION_INT >= AV_VERSION_INT(53,17,0)
 	avformat_close_input(&f);
@@ -720,7 +686,7 @@ ffmpeg_stream_tag(struct input_stream *is)
 #endif
 	mpd_ffmpeg_stream_close(stream);
 
-	return tag;
+	return true;
 }
 
 /**
@@ -745,7 +711,8 @@ static const char *const ffmpeg_suffixes[] = {
 	"ram", "rl2", "rm", "rmvb", "roq", "rpl", "rvc", "shn", "smk", "snd",
 	"sol", "son", "spx", "str", "swf", "tgi", "tgq", "tgv", "thp", "ts",
 	"tsp", "tta", "xa", "xvid", "uv", "uv2", "vb", "vid", "vob", "voc",
-	"vp6", "vmd", "wav", "wma", "wmv", "wsaud", "wsvga", "wv", "wve",
+	"vp6", "vmd", "wav", "webm", "wma", "wmv", "wsaud", "wsvga", "wv",
+	"wve",
 	NULL
 };
 
@@ -809,6 +776,7 @@ static const char *const ffmpeg_mime_types[] = {
 	"video/msvideo",
 	"video/ogg",
 	"video/theora",
+	"video/webm",
 	"video/x-dv",
 	"video/x-flv",
 	"video/x-matroska",
@@ -839,7 +807,7 @@ const struct decoder_plugin ffmpeg_decoder_plugin = {
 	.name = "ffmpeg",
 	.init = ffmpeg_init,
 	.stream_decode = ffmpeg_decode,
-	.stream_tag = ffmpeg_stream_tag,
+	.scan_stream = ffmpeg_scan_stream,
 	.suffixes = ffmpeg_suffixes,
 	.mime_types = ffmpeg_mime_types
 };
