@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2010 The Music Player Daemon Project
+ * Copyright (C) 2003-2011 The Music Player Daemon Project
  * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
@@ -18,12 +18,15 @@
  */
 
 #include "config.h"
+#include "pipe_output_plugin.h"
 #include "output_api.h"
 
 #include <stdio.h>
 #include <errno.h>
 
 struct pipe_output {
+	struct audio_output base;
+
 	char *cmd;
 	FILE *fh;
 };
@@ -37,13 +40,17 @@ pipe_output_quark(void)
 	return g_quark_from_static_string("pipe_output");
 }
 
-static void *
-pipe_output_init(G_GNUC_UNUSED const struct audio_format *audio_format,
-		 const struct config_param *param,
+static struct audio_output *
+pipe_output_init(const struct config_param *param,
 		 GError **error)
 {
 	struct pipe_output *pd = g_new(struct pipe_output, 1);
 
+	if (!ao_base_init(&pd->base, &pipe_output_plugin, param, error)) {
+		g_free(pd);
+		return NULL;
+	}
+
 	pd->cmd = config_dup_block_string(param, "command", NULL);
 	if (pd->cmd == NULL) {
 		g_set_error(error, pipe_output_quark(), 0,
@@ -51,23 +58,25 @@ pipe_output_init(G_GNUC_UNUSED const struct audio_format *audio_format,
 		return NULL;
 	}
 
-	return pd;
+	return &pd->base;
 }
 
 static void
-pipe_output_finish(void *data)
+pipe_output_finish(struct audio_output *ao)
 {
-	struct pipe_output *pd = data;
+	struct pipe_output *pd = (struct pipe_output *)ao;
 
 	g_free(pd->cmd);
+	ao_base_finish(&pd->base);
 	g_free(pd);
 }
 
 static bool
-pipe_output_open(void *data, G_GNUC_UNUSED struct audio_format *audio_format,
+pipe_output_open(struct audio_output *ao,
+		 G_GNUC_UNUSED struct audio_format *audio_format,
 		 G_GNUC_UNUSED GError **error)
 {
-	struct pipe_output *pd = data;
+	struct pipe_output *pd = (struct pipe_output *)ao;
 
 	pd->fh = popen(pd->cmd, "w");
 	if (pd->fh == NULL) {
@@ -81,17 +90,17 @@ pipe_output_open(void *data, G_GNUC_UNUSED struct audio_format *audio_format,
 }
 
 static void
-pipe_output_close(void *data)
+pipe_output_close(struct audio_output *ao)
 {
-	struct pipe_output *pd = data;
+	struct pipe_output *pd = (struct pipe_output *)ao;
 
 	pclose(pd->fh);
 }
 
 static size_t
-pipe_output_play(void *data, const void *chunk, size_t size, GError **error)
+pipe_output_play(struct audio_output *ao, const void *chunk, size_t size, GError **error)
 {
-	struct pipe_output *pd = data;
+	struct pipe_output *pd = (struct pipe_output *)ao;
 	size_t ret;
 
 	ret = fwrite(chunk, 1, size, pd->fh);
