@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2010 The Music Player Daemon Project
+ * Copyright (C) 2003-2011 The Music Player Daemon Project
  * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
@@ -27,6 +27,8 @@
 
 #include <time.h>
 
+struct config_param;
+
 enum audio_output_command {
 	AO_COMMAND_NONE = 0,
 	AO_COMMAND_ENABLE,
@@ -63,12 +65,6 @@ struct audio_output {
 	 */
 	const struct audio_output_plugin *plugin;
 
-	/**
-	 * The plugin's internal data.  It is passed to every plugin
-	 * method.
-	 */
-	void *data;
-
 	/**
 	 * The #mixer object associated with this audio output device.
 	 * May be NULL if none is available, or if software volume is
@@ -109,6 +105,15 @@ struct audio_output {
 	 */
 	bool pause;
 
+	/**
+	 * When this flag is set, the output thread will not do any
+	 * playback.  It will wait until the flag is cleared.
+	 *
+	 * This is used to synchronize the "clear" operation on the
+	 * shared music pipe during the CANCEL command.
+	 */
+	bool allow_play;
+
 	/**
 	 * If not NULL, the device has failed, and this timer is used
 	 * to estimate how long it should stay disabled (unless
@@ -207,6 +212,12 @@ struct audio_output {
 	 */
 	GCond *cond;
 
+	/**
+	 * The player_control object which "owns" this output.  This
+	 * object is needed to signal command completion.
+	 */
+	struct player_control *player_control;
+
 	/**
 	 * The #music_chunk which is currently being played.  All
 	 * chunks before this one may be returned to the
@@ -239,4 +250,20 @@ audio_output_command_is_finished(const struct audio_output *ao)
 	return ao->command == AO_COMMAND_NONE;
 }
 
+struct audio_output *
+audio_output_new(const struct config_param *param,
+		 struct player_control *pc,
+		 GError **error_r);
+
+bool
+ao_base_init(struct audio_output *ao,
+	     const struct audio_output_plugin *plugin,
+	     const struct config_param *param, GError **error_r);
+
+void
+ao_base_finish(struct audio_output *ao);
+
+void
+audio_output_free(struct audio_output *ao);
+
 #endif
