@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2010 The Music Player Daemon Project
+ * Copyright (C) 2003-2011 The Music Player Daemon Project
  * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
@@ -46,6 +46,13 @@ struct queue_item {
 
 	/** when was this item last changed? */
 	uint32_t version;
+
+	/**
+	 * The priority of this item, between 0 and 255.  High
+	 * priority value means that this song gets played first in
+	 * "random" mode.
+	 */
+	uint8_t priority;
 };
 
 /**
@@ -181,6 +188,15 @@ queue_position_to_order(const struct queue *queue, unsigned position)
 	}
 }
 
+G_GNUC_PURE
+static inline uint8_t
+queue_get_priority_at_position(const struct queue *queue, unsigned position)
+{
+	assert(position < queue->length);
+
+	return queue->items[position].priority;
+}
+
 /**
  * Returns the song at the specified position.
  */
@@ -264,9 +280,11 @@ queue_modify_all(struct queue *queue);
  *
  * If a song is not in the database (determined by
  * song_in_database()), it is freed when removed from the queue.
+ *
+ * @param priority the priority of this new queue item
  */
 unsigned
-queue_append(struct queue *queue, struct song *song);
+queue_append(struct queue *queue, struct song *song, uint8_t priority);
 
 /**
  * Swaps two songs, addressed by their position.
@@ -319,6 +337,14 @@ queue_restore_order(struct queue *queue)
 		queue->order[i] = i;
 }
 
+/**
+ * Shuffle the order of items in the specified range, taking their
+ * priorities into account.
+ */
+void
+queue_shuffle_order_range_with_priority(struct queue *queue,
+					unsigned start, unsigned end);
+
 /**
  * Shuffles the virtual order of songs, but does not move them
  * physically.  This is used in random mode.
@@ -341,4 +367,13 @@ queue_shuffle_order_last(struct queue *queue, unsigned start, unsigned end);
 void
 queue_shuffle_range(struct queue *queue, unsigned start, unsigned end);
 
+bool
+queue_set_priority(struct queue *queue, unsigned position,
+		   uint8_t priority, int after_order);
+
+bool
+queue_set_priority_range(struct queue *queue,
+			 unsigned start_position, unsigned end_position,
+			 uint8_t priority, int after_order);
+
 #endif
