@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2010 The Music Player Daemon Project
+ * Copyright (C) 2003-2011 The Music Player Daemon Project
  * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
@@ -20,11 +20,13 @@
 #ifndef MPD_STATE_FILE_H
 #define MPD_STATE_FILE_H
 
-void
-state_file_init(const char *path);
+struct player_control;
 
 void
-state_file_finish(void);
+state_file_init(const char *path, struct player_control *pc);
+
+void
+state_file_finish(struct player_control *pc);
 
 void write_state_file(void);
 
